#include <iostream>
#include <fstream>
#include <map>
#include <set>

#define MOVE     0
#define PICKUP_A 1
#define PICKUP_B 2
#define DROP_A   3
#define DROP_B   4

#define BALLS_A  1
#define BALLS_B  2
#define HOLD     4

int nballs = 0;
int states = 1;
bool nomap = false;
const char *aname[] = { "Move", "PickupA", "PickupB", "DropA", "DropB" };
std::set<int> actions, oset;

bool
compatible( int action, int obs )
{
  if( (action == PICKUP_A) && !(obs&BALLS_A) ) return(false);
  if( (action == PICKUP_B) && !(obs&BALLS_B) ) return(false);
  if( ((action == PICKUP_A) || (action == PICKUP_B)) && (obs&HOLD) ) return(false);
  if( ((action == DROP_A) || (action == DROP_B)) && !(obs&HOLD) ) return(false);
  return(true);
}

void
gen_domain()
{
  for( int o = 0; o < 8; ++o ) oset.insert(o);
  for( int a = 0; a < 5; ++a ) actions.insert(a);
}

const char *
action_name( int action, int nstate )
{
  static char str[64];
  sprintf(str,"%s-q%d",aname[action],nstate);
  return(str);
}

const char *
obs_name( int obs )
{
  static char str[64];
  strcpy(str,"---");
  if( obs&BALLS_A ) str[0] = 'A';
  if( obs&BALLS_B ) str[1] = 'B';
  if( obs&HOLD ) str[2] = 'H';
  return(str);
}

void
emit_action( std::ostream &os, int map, int obs, int state, int action, int nstate )
{
  if( !compatible(action,obs) ) return;
  const char *ostr = obs_name(obs);
  const char *astr = action_name(action,nstate);

  if( map ) {
    os << "  (:action map-" << ostr << "-q" << state << "-" << astr
       << " :precondition (not-obs" << ostr << "-q" << state << ")"
       << " :effect (and (not (not-obs" << ostr << "-q" << state << ")) (obs" << ostr << "-q" << state << "-" << astr << "))" << " )" << std::endl;
  }
  else {
    os << "  (:action app-" << ostr << "-q" << state << "-" << astr << std::endl;
    if( !nomap ) os << "    :precondition (obs" << ostr << "-q" << state << "-" << astr << ")" << std::endl;
    os << "    :effect (and ";

    if( action == MOVE ) {
      os << "(when (and (obs" << ostr << ") (q" << state << ") (at-A)) (and (not (at-A)) (at-B) (not (obs" << ostr << "))";
      if( state != nstate ) os << " (not (q" << state << ")) (q" << nstate << ")";
      os << "))" << std::endl << "                 ";
      os << "(when (and (obs" << ostr << ") (q" << state << ") (at-B)) (and (not (at-B)) (at-A) (not (obs" << ostr << "))";
      if( state != nstate ) os << " (not (q" << state << ")) (q" << nstate << ")";
      os << ")) ))" << std::endl;
    }
    else if( (action >= 1) && (action < 3) ) { // Pickup
      char room = (action==1?'A':'B');
      for( int n = 1; n <= nballs; ++n ) {
        os << "(when (and (obs" << ostr << ") (q" << state << ") (nballs-at-" << room << " n" << n << ")"
           << " (at-" << room << ") (not (holding)))"
           << " (and (nballs-at-" << room << " n" << n-1 << ") (not (nballs-at-" << room << " n" << n << ")) (holding)";
	if( n == 1 ) os << " (not (some-ball-at-" << room << "))";
       	os << " (not (obs" << ostr << "))";
        if( state != nstate ) os << " (not (q" << state << ")) (q" << nstate << ")";
        os << "))";
        if( n < nballs ) os << std::endl << "                 ";
      }
      os << " ))" << std::endl;
    }
    else if( (action >= 3) && (action < 5) ) { // Drop
      char room = (action==3?'A':'B');
      for( int n = 0; n < nballs; ++n ) {
        os << "(when (and (obs" << ostr << ") (q" << state << ") (nballs-at-" << room << " n" << n << ")"
           << " (at-" << room << ") (holding))"
           << " (and (nballs-at-" << room << " n" << n+1 << ") (not (nballs-at-" << room << " n" << n << ")) (not (holding))";
	if( n == 0 ) os << " (some-ball-at-" << room << ")";
       	os << " (not (obs" << ostr << "))";
        if( state != nstate ) os << " (not (q" << state << ")) (q" << nstate << ")";
        os << "))";
        if( n < nballs-1 ) os << std::endl << "                 ";
      }
      os << " ))" << std::endl;
    }
  }
}

void
emit_zoom( std::ostream &os )
{
  os << "  (:action zoom" << std::endl
     << "    :effect (and ";

  os << "(when (and (some-ball-at-A) (some-ball-at-B) (holding)) (obsABH))"
     << std::endl <<  "                 "
     << "(when (and (some-ball-at-A) (some-ball-at-B) (not (holding))) (obsAB-))"
     << std::endl <<  "                 "
     << "(when (and (some-ball-at-A) (not (some-ball-at-B)) (holding)) (obsA-H))"
     << std::endl <<  "                 "
     << "(when (and (some-ball-at-A) (not (some-ball-at-B)) (not (holding))) (obsA--))"
     << std::endl <<  "                 "
     << "(when (and (not (some-ball-at-A)) (some-ball-at-B) (holding)) (obs-BH))"
     << std::endl <<  "                 "
     << "(when (and (not (some-ball-at-A)) (some-ball-at-B) (not (holding))) (obs-B-))"
     << std::endl <<  "                 "
     << "(when (and (not (some-ball-at-A)) (not (some-ball-at-B)) (holding)) (obs--H))"
     << std::endl <<  "                 "
     << "(when (and (not (some-ball-at-A)) (not (some-ball-at-B)) (not (holding))) (obs---))";

  os << " ))" << std::endl;
}

void
emit_preamble( std::ostream &os )
{
  os << "(define (domain gripper)" << std::endl
     << "  (:predicates (at-A) (at-B) (holding) (some-ball-at-A) (some-ball-at-B) (nballs-at-A ?p) (nballs-at-B ?p)" << std::endl;

  for( std::set<int>::const_iterator ot = oset.begin(); ot != oset.end(); ++ot ) {
    const char *ostr = obs_name(*ot);
    os << "               (obs" << ostr << ")";
    if( !nomap ) {
      for( int state = 0; state < states; ++state ) {
        os << " (not-obs" << ostr << "-q" << state << ")";
        for( int action = 0; action < 5; ++action ) {
          for( int nstate = 0; nstate < states; ++nstate ) {
            const char *astr = action_name(action,nstate);
            os << " (obs" << ostr << "-q" << state << "-" << astr << ")";
          }
        }
      }
    }
    os << std::endl;
  }

  for( int state = 0; state < states; ++state )
    os << "               (q" << state << ")" << std::endl;

  os << "  )" << std::endl;
  os << "  (:constants";
  for( int n = 0; n <= nballs; ++n ) os << " n" << n;
  os << ")" << std::endl;
}

void
emit_domain( std::ostream &os )
{
  emit_preamble(os);
  for( int action = 0; action < 5; ++action ) {
    for( std::set<int>::const_iterator ot = oset.begin(); ot != oset.end(); ++ot ) {
      for( int state = 0; state < states; ++state ) {
        for( int nstate = 0; nstate < states; ++nstate ) {
          if( !nomap ) emit_action(os,true,*ot,state,action,nstate);
          emit_action(os,false,*ot,state,action,nstate);
        }
      }
    }
  }
  emit_zoom(os);
  os << ")" << std::endl;
}

void
emit_problem( std::ostream &os )
{
  os << "(define (problem p)" << std::endl
     << "  (:domain gripper)" << std::endl
     << "  (:init" << std::endl
     << "    (q0)" << std::endl
     << "    (at-A)" << std::endl
     << "    (nballs-at-B n" << nballs << ")" << std::endl
     << "    (some-ball-at-B)" << std::endl
     << "    (nballs-at-A n0)" << std::endl;

  if( !nomap ) {
    for( std::set<int>::const_iterator ot = oset.begin(); ot != oset.end(); ++ot ) {
      const char *ostr = obs_name(*ot);
      for( int state = 0; state < states; ++state ) {
        os << "    (not-obs" << ostr << "-q" << state << ")" << std::endl;
      }
    }
  }

  os << "  )" << std::endl
     << "  (:goal (nballs-at-A n" << nballs << "))" << std::endl
     << ")" << std::endl;
}

int
main( int argc, char **argv )
{
  if( argc < 1 ) {
    std::cout << "usage: " << argv[0] << " [--nomap] [--states <n>] [-p|-d] <nballs>" << std::endl;
    return(0);
  }

  bool problem = false, domain = false;
  for( ++argv, --argc; (argc > 0) && (**argv == '-'); ++argv, --argc ) {
    if( !strcmp(*argv,"--nomap") ) {
      nomap = true;
    }
    else if( !strcmp(*argv,"--states") ) {
      states = atoi(*++argv);
      --argc;
    }
    else if( !strcmp(*argv,"-p") ) {
      problem = true;
    }
    else if( !strcmp(*argv,"-d") ) {
      domain = true;
    }
  }
  
  nballs = atoi(argv[0]);
  gen_domain();
  if( domain ) emit_domain(std::cout);
  if( problem ) emit_problem(std::cout);

  return(0);
}

