#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <iostream>
#include <map>
#include <set>

#define NORTH   0
#define EAST    1
#define SOUTH   2
#define WEST    3

#define FORWARD 0
#define TURNCW  1
#define TURNCCW 2

#define A       1
#define B       2
#define C       4
#define D       8
#define WALL    16

int xdim = 0;
int ydim = 0;
int states = 1;
bool nomap = false;
bool unknown_init = false;
const char *aname[] = { "Forwd", "TurnR", "TurnL" };
std::set<int> actions;
std::map<std::pair<std::pair<int,int>,int>,int> omap;
std::set<int> oset;

bool
compatible( int action, int obs )
{
  return(true);
}

void
gen_domain()
{
  for( int p = 0; p < xdim; ++p ) {
    for( int h = 0; h < 4; ++h ) {
      int obs = 0;
      if( p == 0 ) {
        obs += A;
        if( (h == NORTH) || (h == WEST) ) obs += WALL;
	if( (h == SOUTH) && (ydim == 1) ) obs += WALL;
	if( (h == EAST) && (xdim == 1) ) obs += WALL;
      }
      else if( (xdim > 1) && (p == xdim-1) ) {
        obs = B;
        if( (h == NORTH) || (h == EAST) ) obs += WALL;
	if( (h == SOUTH) && (ydim == 1) ) obs += WALL;
      }
      else {
        if( (h == NORTH) || (h == SOUTH) ) obs += WALL;
      }
      omap.insert(std::make_pair(std::make_pair(std::make_pair(p+1,1),h),obs));
      oset.insert(obs);
      if( ydim > 1 ) {
        obs = 0;
        if( p == 0 ) {
          obs += D;
          if( (h == SOUTH) || (h == WEST) ) obs += WALL;
	  if( (h == EAST) && (xdim == 1) ) obs += WALL;
        }
        else if( (xdim > 1) && (p == xdim-1) ) {
          obs = C;
          if( (h == SOUTH) || (h == EAST) ) obs += WALL;
        }
        else {
          if( (h == NORTH) || (h == SOUTH) ) obs += WALL;
        }
        omap.insert(std::make_pair(std::make_pair(std::make_pair(p+1,ydim),h),obs));
        oset.insert(obs);
      }
    }
  }
  for( int q = 0; q < ydim; ++q ) {
    for( int h = 0; h < 4; ++h ) {
      int obs = 0;
      if( q == 0 ) {
        obs += A;
        if( (h == NORTH) || (h == WEST) ) obs += WALL;
	if( (h == SOUTH) && (ydim == 1) ) obs += WALL;
	if( (h == EAST) && (xdim == 1) ) obs += WALL;
      }
      else if( (ydim > 1) && (q == ydim-1) ) {
        obs = D;
        if( (h == SOUTH) || (h == WEST) ) obs += WALL;
	if( (h == EAST) && (xdim == 1) ) obs += WALL;
      }
      else {
        if( (h == EAST) || (h == WEST) ) obs += WALL;
      }
      omap.insert(std::make_pair(std::make_pair(std::make_pair(1,q+1),h),obs));
      oset.insert(obs);
      if( xdim > 1 ) {
        obs = 0;
        if( q == 0 ) {
          obs += B;
          if( (h == NORTH) || (h == EAST) ) obs += WALL;
	  if( (h == SOUTH) && (ydim == 1) ) obs += WALL;
        }
        else if( (ydim > 1) && (q == ydim-1) ) {
          obs = C;
          if( (h == SOUTH) || (h == EAST) ) obs += WALL;
        }
        else {
          if( (h == EAST) || (h == WEST) ) obs += WALL;
        }
        omap.insert(std::make_pair(std::make_pair(std::make_pair(xdim,q+1),h),obs));
        oset.insert(obs);
      }
    }
  }
  actions.insert(0);
  actions.insert(1);
  actions.insert(2);
}

const char *
action_name( int action, int nstate )
{
  static char str[64];
  sprintf(str,"%s-q%d",aname[action],nstate);
  return(str);
}

const char *
obs_name( int obs )
{
  static char str[64];
  sprintf(str,"--");
  if( obs&A ) str[0] = 'A';
  if( obs&B ) str[0] = 'B';
  if( obs&C ) str[0] = 'C';
  if( obs&D ) str[0] = 'D';
  if( obs&WALL ) str[1] = 'W';
  return(str);
}

void
emit_action( std::ostream &os, int map, int obs, int state, int action, int nstate )
{
  if( (action == FORWARD) && (obs&WALL) ) return;

  const char *ostr = obs_name(obs);
  const char *astr = action_name(action,nstate);

  if( map ) {
    os << "  (:action map-" << ostr << "-q" << state << "-" << astr
       << " :precondition (not-obs" << ostr << "-q" << state << ")"
       << " :effect (and (not (not-obs" << ostr << "-q" << state << ")) (obs" << ostr << "-q" << state << "-" << astr << "))" << " )" << std::endl;
  }
  else {
    os << "  (:action app-" << ostr << "-q" << state << "-" << astr << std::endl;
    if( !nomap ) os << "    :precondition (obs" << ostr << "-q" << state << "-" << astr << ")" << std::endl;

    os << "    :effect (and ";

    if( action == FORWARD ) {
      for( int p = 2; p <= ydim; ++p ) {
        os << "(when (and (obs" << ostr << ") (q" << state << ") (north) (x-agent p1) (y-agent p" << p << "))"
           << " (and (not (y-agent p" << p << ")) (y-agent p" << p-1 << ")" << " (not (obs" << ostr << "))";
        if( state != nstate ) os << " (not (q" << state << ")) (q" << nstate << ")";
        os << "))" << std::endl << "                 ";
        os << "(when (and (obs" << ostr << ") (q" << state << ") (north) (x-agent p" << xdim << ") (y-agent p" << p << "))"
           << " (and (not (y-agent p" << p << ")) (y-agent p" << p-1 << ")" << " (not (obs" << ostr << "))";
        if( state != nstate ) os << " (not (q" << state << ")) (q" << nstate << ")";
        os << "))" << std::endl << "                 ";
      }
      for( int p = 1; p <= ydim-1; ++p ) {
        os << "(when (and (obs" << ostr << ") (q" << state << ") (south) (x-agent p1) (y-agent p" << p << "))"
           << " (and (not (y-agent p" << p << ")) (y-agent p" << p+1 << ")" << " (not (obs" << ostr << "))";
        if( state != nstate ) os << " (not (q" << state << ")) (q" << nstate << ")";
        os << "))" << std::endl << "                 ";
        os << "(when (and (obs" << ostr << ") (q" << state << ") (south) (x-agent p" << xdim << ") (y-agent p" << p << "))"
           << " (and (not (y-agent p" << p << ")) (y-agent p" << p+1 << ")" << " (not (obs" << ostr << "))";
        if( state != nstate ) os << " (not (q" << state << ")) (q" << nstate << ")";
        os << "))" << std::endl << "                 ";
      }
      for( int p = 1; p <= xdim-1; ++p ) {
        os << "(when (and (obs" << ostr << ") (q" << state << ") (east) (x-agent p" << p << ") (y-agent p1))"
           << " (and (not (x-agent p" << p << ")) (x-agent p" << p+1 << ")" << " (not (obs" << ostr << "))";
        if( state != nstate ) os << " (not (q" << state << ")) (q" << nstate << ")";
        os << "))" << std::endl << "                 ";
        os << "(when (and (obs" << ostr << ") (q" << state << ") (east) (x-agent p" << p << ") (y-agent p" << ydim << "))"
           << " (and (not (x-agent p" << p << ")) (x-agent p" << p+1 << ")" << " (not (obs" << ostr << "))";
        if( state != nstate ) os << " (not (q" << state << ")) (q" << nstate << ")";
        os << "))" << std::endl << "                 ";
      }
      for( int p = 2; p <= xdim; ++p ) {
        os << "(when (and (obs" << ostr << ") (q" << state << ") (west) (x-agent p" << p << ") (y-agent p1))"
           << " (and (not (x-agent p" << p << ")) (x-agent p" << p-1 << ")" << " (not (obs" << ostr << "))";
        if( state != nstate ) os << " (not (q" << state << ")) (q" << nstate << ")";
        os << "))" << std::endl << "                 ";
        os << "(when (and (obs" << ostr << ") (q" << state << ") (west) (x-agent p" << p << ") (y-agent p" << ydim << "))"
           << " (and (not (x-agent p" << p << ")) (x-agent p" << p-1 << ")" << " (not (obs" << ostr << "))";
        if( state != nstate ) os << " (not (q" << state << ")) (q" << nstate << ")";
        os << "))" << std::endl << "                 ";
      }
      os << std::endl;
    }
    else {
      os << "(when (and (obs" << ostr << ") (q" << state << ") (north))"
         << " (and (not (north)) " << (action==TURNCW?"(east)":"(west)") << " (not (obs" << ostr << "))";
      if( state != nstate ) os << " (not (q" << state << ")) (q" << nstate << ")";
      os << "))" << std::endl << "                 ";
      os << "(when (and (obs" << ostr << ") (q" << state << ") (east))"
         << " (and (not (east)) " << (action==TURNCW?"(south)":"(north)") << " (not (obs" << ostr << "))";
      if( state != nstate ) os << " (not (q" << state << ")) (q" << nstate << ")";
      os << "))" << std::endl << "                 ";
      os << "(when (and (obs" << ostr << ") (q" << state << ") (south))"
         << " (and (not (south)) " << (action==TURNCW?"(west)":"(east)") << " (not (obs" << ostr << "))";
      if( state != nstate ) os << " (not (q" << state << ")) (q" << nstate << ")";
      os << "))" << std::endl << "                 ";
      os << "(when (and (obs" << ostr << ") (q" << state << ") (west))"
         << " (and (not (west)) " << (action==TURNCW?"(north)":"(south)") << " (not (obs" << ostr << "))";
      if( state != nstate ) os << " (not (q" << state << ")) (q" << nstate << ")";
      os << "))" << std::endl;
    }

    os << " ))" << std::endl;
  }
}

void
emit_obs_action( std::ostream &os )
{
  os << "  (:action zoom" << std::endl
     << "    :effect (and ";

  for( std::map<std::pair<std::pair<int,int>,int>,int>::const_iterator ot = omap.begin(); ot != omap.end(); ++ot ) {
    const char *ostr = obs_name((*ot).second);
    os << "(when (and (x-agent p" << (*ot).first.first.first << ") (y-agent p" << (*ot).first.first.second << ")";
    switch( (*ot).first.second ) {
    case 0: os << " (north)) (obs" << ostr << "))" << std::endl << "                 "; break;
    case 1: os << " (east)) (obs" << ostr << "))" << std::endl << "                 "; break;
    case 2: os << " (south)) (obs" << ostr << "))" << std::endl << "                 "; break;
    case 3: os << " (west)) (obs" << ostr << "))" << std::endl << "                 "; break;
    }
  }

  os << "(when (and (x-agent p1) (y-agent p1)) (visited-A))";
  if( xdim > 1 ) os << std::endl << "                 (when (and (x-agent p" << xdim << ") (y-agent p1)) (visited-B))";
  if( (xdim > 1) && (ydim > 1) ) os << std::endl << "                 (when (and (x-agent p" << xdim << ") (y-agent p" << ydim << ")) (visited-C))";
  if( ydim > 1 ) os << std::endl << "                 (when (and (x-agent p1) (y-agent p" << ydim << ")) (visited-D))";

  os << " ))" << std::endl;
}

void
emit_preamble( std::ostream &os )
{
  os << "(define (domain hall-rotate)" << std::endl
     << "  (:predicates (x-agent ?p) (y-agent ?p)" << std::endl
     << "               (north) (east) (south) (west)" << std::endl
     << "               (visited-A) (visited-B) (visited-C) (visited-D)" << std::endl;
  
  for( std::set<int>::const_iterator ot = oset.begin(); ot != oset.end(); ++ot ) {
    const char *ostr = obs_name(*ot);
    os << "               (obs" << ostr << ")" << std::endl;
    if( !nomap ) {
      for( int state = 0; state < states; ++state ) {
        os << "               (not-obs" << ostr << "-q" << state << ")" << std::endl;
        for( std::set<int>::const_iterator action = actions.begin(); action != actions.end(); ++action ) {
          for( int nstate = 0; nstate < states; ++nstate ) {
            const char *astr = action_name(*action,nstate);
            os << "               (obs" << ostr << "-q" << state << "-" << astr << ")" << std::endl;
          }
        }
      }
    }
  }

  for( int state = 0; state < states; ++state ) os << "               (q" << state << ")" << std::endl;
  
  os << "  )" << std::endl << "  (:constants";
  int maxp = (xdim>ydim?xdim:ydim);
  for( int p = 1; p <= maxp; ++p ) os << " p" << p;
  os << ")" << std::endl;
}

void
emit_domain( std::ostream &os )
{
  emit_preamble(os);
  for( std::set<int>::const_iterator at = actions.begin(); at != actions.end(); ++at ) {
    for( std::set<int>::const_iterator ot = oset.begin(); ot != oset.end(); ++ot ) {
      if( !compatible(*at,*ot) ) continue;
      for( int state = 0; state < states; ++state ) {
        for( int nstate = 0; nstate < states; ++nstate ) {
          if( !nomap ) emit_action(os,true,*ot,state,*at,nstate);
          emit_action(os,false,*ot,state,*at,nstate);
        }
      }
    }
  }
  emit_obs_action(os);
  os << ")" << std::endl;
}

void
emit_problem( std::ostream &os )
{
  os << "(define (problem p)" << std::endl
     << "  (:domain hall-rotate)" << std::endl
     << "  (:init" << std::endl
     << "    (q0) (north)" << std::endl
     << "    (y-agent p1)" << std::endl;

  if( !unknown_init || (xdim <= 2) ) {
    os << "    (x-agent p1)" << std::endl;
  }
  else {
    os << "    (oneof";
    for( int p = 1; p <= xdim-1; ++p ) os << " (x-agent p" << p << ")";
    os << ")" << std::endl;
  }

  if( !nomap ) {
    for( std::set<int>::const_iterator ot = oset.begin(); ot != oset.end(); ++ot ) {
      for( int state = 0; state < states; ++state ) {
        const char *ostr = obs_name(*ot);
        os << "    (not-obs" << ostr << "-q" << state << ")" << std::endl;
      }
    }
  }

  os << "  )" << std::endl << "  (:goal (and";
  if( xdim > 1 ) os << " (x-agent p1)";
  if( ydim > 1 ) os << " (y-agent p1)";
  os << " (visited-A)";
  if( xdim > 1 ) os << " (visited-B)";
  if( ydim > 1 ) os << " (visited-D)";
  if( (xdim > 1) && (ydim > 1) ) os << " (visited-C)";
  os << "))" << std::endl << ")" << std::endl;
}

int
main( int argc, char **argv )
{
  if( argc < 4 ) {
    std::cout << "usage: " << argv[0] << " [--nomap] [--states <n>] [--unknown-init] [-p|-d] <rows> <cols>" << std::endl;
    return(0);
  }

  bool problem = false, domain = false;
  for( ++argv, --argc; argc > 2; ++argv, --argc ) {
    if( !strcmp(*argv,"--nomap") ) {
      nomap = true;
    }
    else if( !strcmp(*argv,"--states") ) {
      states = atoi(*++argv);
      --argc;
    }
    else if( !strcmp(*argv,"--unknown-init") ) {
      unknown_init = true;
    }
    else if( !strcmp(*argv,"-p") ) {
      problem = true;
    }
    else if( !strcmp(*argv,"-d") ) {
      domain = true;
    }
  }
  
  ydim = atoi(argv[0]);
  xdim = atoi(argv[1]);
  gen_domain();
  if( domain ) emit_domain(std::cout);
  if( problem ) emit_problem(std::cout);
  return(0);
}

