#include <iostream>
#include <fstream>
#include <map>
#include <set>
#include <string.h>
#include <stdlib.h>

using namespace std;

#define UP      0
#define RIGHT   1
#define DOWN    2
#define LEFT    3
#define LOCK    4
#define CLEAR   0
#define BLOCK   1
#define GREEN   2
#define TABLE   4

const char *rfile = 0;
const char *wfile = 0;
int nblocks = 0;
int dim = 0, xdim = 0, ydim = 0;
int *grid = 0;
int states = 1;
bool nomap = false, costs = false, init_and = false;
const char *aname[] = { "Up", "Right", "Down", "Left", "Lock" };
set<int> actions, oset;

bool compatible(int action, int obs) {
    if( action == LOCK ) return false; // test
    if( (action == UP) && (obs == CLEAR) ) return false;
    if( (action == LOCK) && (obs != 2) && (obs != 6) ) return false;
    if( (action != LOCK) && ((obs == 2) || (obs == 6)) ) return false;
    if( (action == LEFT) &&  !(obs&TABLE) ) return false;
    if( (action == RIGHT) &&  !(obs&TABLE) ) return false;
    return true;
}

void read_file(const char *file) {
    oset.insert(CLEAR);
    oset.insert(CLEAR+TABLE);
    oset.insert(BLOCK);
    oset.insert(BLOCK+TABLE);
    oset.insert(GREEN);
    oset.insert(GREEN+TABLE);

    int ntowers = 0;
    ifstream ifs(file,ifstream::in);
    ifs >> nblocks >> ydim >> ntowers;

    grid = new int[nblocks*ydim];
    bzero(grid,nblocks*ydim*sizeof(int));

    xdim = 0;
    for( int t = 0; t < ntowers; ++t ) {
        int p, blks;
        ifs >> p >> blks;
        for( int q = 0; q < blks; ++q ) grid[(p-1)*ydim+q] = BLOCK;
        xdim = p > xdim ? p : xdim;
    }
    ifs.close();
    dim =  xdim > ydim ? xdim : ydim;
}

void write_file(const char *file) {
    ofstream ofs(file,ofstream::out);
    ofs << nblocks << " " << ydim << " ";
    int ntowers = 0;
    for( int p = 0; p < xdim; ++p ) ntowers += grid[p*ydim] == BLOCK ? 1 : 0;
    ofs << ntowers;
    for( int p = 0; p < xdim; ++p ) {
        if( grid[p*ydim] == BLOCK ) {
            int blks = 0;
            for( int q = 0; q < ydim; ++q ) blks += grid[p*ydim+q] == BLOCK ? 1 : 0;
            ofs << " " << 1+p << " " << blks;
        }
    }
    ofs << endl;
    ofs.close();
}

void random_domain() {
    oset.insert(CLEAR);
    oset.insert(CLEAR+TABLE);
    oset.insert(BLOCK);
    oset.insert(BLOCK+TABLE);
    oset.insert(GREEN);
    oset.insert(GREEN+TABLE);

    grid = new int[nblocks*ydim];
    bzero(grid,nblocks*ydim*sizeof(int));

    int n = nblocks;
    int *done = new int[nblocks];
    bzero(done,nblocks*sizeof(int));

    int p = 0, block = 0;
    while( n > 0 ) {
        int m = 0, min = ydim < n ? ydim : n;
        if( (done[p] == 0) && (drand48() > .35) ) {
            m = 1+(lrand48()%min);
            done[p] = 1;
            n -= m;
            for( int q = 0; q < m; ++q ) {
                grid[p*ydim+q] = BLOCK;
                ++block;
            }
        }
        ++p;
        if( p == nblocks ) { p = 0; xdim = nblocks; }
    }
    delete[] done;

    if( xdim == 0 ) xdim = p;
    dim = xdim > ydim ? xdim : ydim;
}

const char* action_name(int action, int nstate) {
    static char str[64];
    sprintf(str,"%s-q%d",aname[action],nstate);
    return str;
}

const char* obs_name(int obs) {
    static char str[64];
    strcpy(str,"--");
    if( obs&0x4 ) str[0] = 'T';
    if( (obs&~0x4) == CLEAR ) str[1] = 'C';
    if( (obs&~0x4) == BLOCK ) str[1] = 'B';
    if( (obs&~0x4) == GREEN ) str[1] = 'G';
    return str;
}

void emit_action(ostream &os, int map, int obs, int state, int action, int nstate) {
    if( !compatible(action,obs) ) return;
    const char *ostr = obs_name(obs);
    const char *astr = action_name(action,nstate);

    if( map ) {
        os << "  (:action map-" << ostr << "-q" << state << "-" << astr << endl
           << "    :precondition (not-obs" << ostr << "-q" << state << ")" << endl
           << "    :effect (and (not (not-obs" << ostr << "-q" << state << "))" << endl
           << "                 (obs" << ostr << "-q" << state << "-" << astr << ")" << endl
           << "                 ";
    } else {
        os << "  (:action app-" << ostr << "-q" << state << "-" << astr << endl;
        if( !nomap ) os << "    :precondition (obs" << ostr << "-q" << state << "-" << astr << ")" << endl;
        os << "    :effect (and ";
    }

    if( action == LOCK ) {
        os << "(when (and (obs" << ostr << ") (q" << state << ")) (acquired))";
    } else {
        int dim = xdim;
        if( (action == UP) || (action == DOWN) ) dim = ydim;
        if( (action == UP) || (action == RIGHT) ) --dim;

        for( int p = 0; p < dim; ++p ) {
            bool something = false;
            if( action == UP ) {
                os << "(when (and (obs" << ostr << ") (q" << state << ") (y-pos p" << 1+p << "))"
                   << " (and (not (y-pos p" << 1+p << ")) (y-pos p" << 2+p << ")";
                something = true;
            } else if( action == RIGHT ) {
                os << "(when (and (obs" << ostr << ") (q" << state << ") (x-pos p" << 1+p << "))"
                   << " (and (not (x-pos p" << 1+p << ")) (x-pos p" << 2+p << ")";
                something = true;
            } else if( (action == DOWN) && (p > 0) ) {
                os << "(when (and (obs" << ostr << ") (q" << state << ") (y-pos p" << 1+p << "))"
                   << " (and (not (y-pos p" << 1+p << ")) (y-pos p" << p << ")";
                something = true;
            } else if( (action == LEFT) && (p > 0) ) {
                os << "(when (and (obs" << ostr << ") (q" << state << ") (x-pos p" << 1+p << "))"
                   << " (and (not (x-pos p" << 1+p << ")) (x-pos p" << p << ")";
                something = true;
            }

            if( something ) {
                os << " (not (obs" << ostr << "))";
                if( state != nstate ) os << " (not (q" << state << ")) (q" << nstate << ")";
                os << "))";
                if( p < dim-1 ) os << endl << "                 ";
            }
        }
    }
  
    if( costs ) os << endl << "                 (increase (total-cost) 1)";
    os << " ))" << endl;
}

void emit_zoom(ostream &os) {
    os << "  (:action zoom" << endl
       << "    :effect (and ";

    for( int p = 0; p < xdim; ++p ) {
        os << "(when (and (x-pos p" << 1+p << ") (y-pos p1) (not (block p" << 1+p << " p1))) (obsTC))"
           << endl <<  "                 ";
        os << "(when (and (x-pos p" << 1+p << ") (y-pos p1) (block p" << 1+p << " p1)" << " (not (green p" << 1+p << " p1))) (obsTB))"
           << endl <<  "                 ";
        os << "(when (and (x-pos p" << 1+p << ") (y-pos p1) (green p" << 1+p << " p1)) (obsTG))"
           << endl <<  "                 ";
        for( int q = 1; q < ydim; ++q ) {
            os << "(when (and (x-pos p" << 1+p << ") (y-pos p" << 1+q << ") (not (block p" << 1+p << " p" << 1+q << "))) (obs-C))"
               << endl <<  "                 ";
            os << "(when (and (x-pos p" << 1+p << ") (y-pos p" << 1+q << ") (block p" << 1+p << " p" << 1+q << ")"
               << " (not (green p" << 1+p << " p" << 1+q << "))) (obs-B))"
               << endl <<  "                 ";
            os << "(when (and (x-pos p" << 1+p << ") (y-pos p" << 1+q << ") (green p" << 1+p << " p" << 1+q << ")) (obs-G))";
            if( (q < ydim-1) || (p < xdim-1) ) os << endl << "                 ";
        }
    }

    os << " ))" << endl;
}

void emit_preamble(ostream &os) {
    os << "(define (domain marker)" << endl;

    os << "  (:constants";
    for( int p = 0; p < dim; ++p ) os << " p" << 1+p;
    os << ")" << endl;

    os << "  (:predicates (x-pos ?p) (y-pos ?p) (block ?p ?q) (green ?p ?q) (acquired)" << endl;

    for( set<int>::const_iterator ot = oset.begin(); ot != oset.end(); ++ot ) {
        const char *ostr = obs_name(*ot);
        os << "               (obs" << ostr << ")";
        if( !nomap ) {
            for( int state = 0; state < states; ++state ) {
                os << " (not-obs" << ostr << "-q" << state << ")";
                for( int action = 0; action < 5; ++action ) {
                    for( int nstate = 0; nstate < states; ++nstate ) {
                        const char *astr = action_name(action,nstate);
                        os << " (obs" << ostr << "-q" << state << "-" << astr << ")";
                    }
                }
            }
        }
        os << endl;
    }

    for( int state = 0; state < states; ++state )
        os << "               (q" << state << ")" << endl;
    os << "  )" << endl;

    if( costs ) os << "  (:functions (total-cost) number)" << endl;
}

void emit_domain(ostream &os) {
    emit_preamble(os);
    for( int action = 0; action < 5; ++action ) {
        for( set<int>::const_iterator ot = oset.begin(); ot != oset.end(); ++ot ) {
            for( int state = 0; state < states; ++state ) {
                for( int nstate = 0; nstate < states; ++nstate ) {
                    if( !nomap ) emit_action(os,true,*ot,state,action,nstate);
                    emit_action(os,false,*ot,state,action,nstate);
                }
            }
        }
    }
    emit_zoom(os);
    os << ")" << endl;
}

void emit_problem(ostream &os) {
    os << "(define (problem p)" << endl
       << "  (:domain marker)" << endl
       << "  (:init";
    if( init_and ) os << " (and";
    os << endl
       << "    (q0)" << endl
       << "    (x-pos p1)" << endl
       << "    (y-pos p1)" << endl;

    set<pair<int,int> > blocks;
    for( int p = 0; p < xdim; ++p ) {
        for( int q = 0; q < ydim; ++q ) {
            if( grid[p*ydim+q] != CLEAR ) {
                os << "    (block p" << 1+p << " p" << 1+q << ")" << endl;
	        blocks.insert(make_pair(1+p,1+q));
            }
        }
    }

    os << "    (oneof";
    for( set<pair<int,int> >::const_iterator it = blocks.begin(); it != blocks.end(); ++it )
        os << " (green p" << (*it).first << " p" << (*it).second << ")";
    os << ")" << endl;

    if( !nomap ) {
        for( set<int>::const_iterator ot = oset.begin(); ot != oset.end(); ++ot ) {
            const char *ostr = obs_name(*ot);
            for( int state = 0; state < states; ++state ) {
                os << "    (not-obs" << ostr << "-q" << state << ")" << endl;
            }
        }
    }

    os << "  )";
    if( init_and ) os << ")";
    os << endl
       << "  (:goal (or (obsTG) (obs-G)))" << endl;
    if( costs ) os << "  (:metric minimize (total-cost))" << endl;
    os << ")" << endl;
}

int main(int argc, char **argv) {
    if( argc < 2 ) {
        cout << "usage: " << argv[0] << " [--read <file>] [--write <file>] [--costs] [--init-and] [--nomap] [--states <n>] [--seed <n>] [-p|-d] <blocks> <max-height>" << endl;
        return 0;
    }

    bool problem = false, domain = false;
    for( ++argv, --argc; (argc > 0) && (**argv == '-'); ++argv, --argc ) {
        if( !strcmp(*argv,"--read") ) {
            rfile = *++argv;
            --argc;
        } else if( !strcmp(*argv,"--write") ) {
            wfile = *++argv;
            --argc;
        } else if( !strcmp(*argv,"--costs") ) {
            costs = true;
        } else if( !strcmp(*argv,"--init-and") ) {
            init_and = true;
        } else if( !strcmp(*argv,"--nomap") ) {
            nomap = true;
        } else if( !strcmp(*argv,"--states") ) {
            states = atoi(*++argv);
            --argc;
        } else if( !strcmp(*argv,"--seed") ) {
            unsigned short s[3];
            int seed = atoi(*++argv);
            s[0] = s[1] = s[2] = seed;
            srand48(seed);
            seed48(s);
            --argc;
        } else if( !strcmp(*argv,"-p") ) {
            problem = true;
        } else if( !strcmp(*argv,"-d") ) {
            domain = true;
        }
    }
  
    if( rfile == 0 ) {
        nblocks = atoi(argv[0]);
        ydim = atoi(argv[1]);
        random_domain();
    } else {
        read_file(rfile);
    }

    if( domain ) emit_domain(cout);
    if( problem ) emit_problem(cout);
    if( wfile != 0 ) write_file(wfile);

    return 0;
}

